use std::io;
use std::time::{Duration, Instant};

use crossterm::event;
use crossterm::event::{Event, KeyCode};
use tui::backend::Backend;
use tui::style::*;
use tui::text::*;
use tui::widgets::*;
use tui::{Frame, Terminal};

use sea_battle_backend::data::{BotDescription, BotType, PlayConfiguration};

use crate::constants::{HIGHLIGHT_COLOR, TICK_RATE};
use crate::ui_screens::utils::centered_rect_size;
use crate::ui_screens::ScreenResult;

struct SelectPlayModeScreen {
    state: ListState,
    curr_selection: usize,
    types: Vec<BotDescription>,
}

pub fn select_bot_type<B: Backend>(
    terminal: &mut Terminal<B>,
) -> io::Result<ScreenResult<BotType>> {
    let types = PlayConfiguration::default().bot_types;
    let mut model = SelectPlayModeScreen {
        state: Default::default(),
        curr_selection: types.len() - 1,
        types,
    };

    let mut last_tick = Instant::now();
    loop {
        model.state.select(Some(model.curr_selection));
        terminal.draw(|f| ui(f, &mut model))?;

        let timeout = TICK_RATE
            .checked_sub(last_tick.elapsed())
            .unwrap_or_else(|| Duration::from_secs(0));

        if event::poll(timeout)? {
            if let Event::Key(key) = event::read()? {
                match key.code {
                    KeyCode::Char('q') => return Ok(ScreenResult::Canceled),
                    KeyCode::Enter => {
                        return Ok(ScreenResult::Ok(model.types[model.curr_selection].r#type));
                    }
                    KeyCode::Down => model.curr_selection += 1,
                    KeyCode::Up => model.curr_selection += model.types.len() - 1,
                    _ => {}
                }

                model.curr_selection %= model.types.len();
            }
        }
        if last_tick.elapsed() >= TICK_RATE {
            last_tick = Instant::now();
        }
    }
}

fn ui<B: Backend>(f: &mut Frame<B>, model: &mut SelectPlayModeScreen) {
    let area = centered_rect_size(60, model.types.len() as u16 * 2 + 2, &f.size());

    // Create a List from all list items and highlight the currently selected one
    let items = model
        .types
        .iter()
        .map(|bot| {
            ListItem::new(vec![
                Spans::from(bot.name),
                Spans::from(Span::styled(
                    bot.description,
                    Style::default().add_modifier(Modifier::ITALIC),
                )),
            ])
        })
        .collect::<Vec<_>>();
    let items = List::new(items)
        .block(
            Block::default()
                .title("Select bot type")
                .borders(Borders::ALL),
        )
        .highlight_style(
            Style::default()
                .fg(HIGHLIGHT_COLOR)
                .add_modifier(Modifier::BOLD),
        )
        .highlight_symbol(">> ");

    f.render_stateful_widget(items, area, &mut model.state);
}
