use std::io;
use std::time::{Duration, Instant};

use crate::constants::{HIGHLIGHT_COLOR, TICK_RATE};
use crate::ui_screens::utils::centered_rect_size;
use crossterm::event;
use crossterm::event::{Event, KeyCode};
use tui::backend::Backend;
use tui::style::*;
use tui::text::Text;
use tui::widgets::*;
use tui::{Frame, Terminal};

#[derive(Debug, Eq, PartialEq, Copy, Clone, Default)]
pub enum SelectPlayModeResult {
    #[default]
    PlayAgainstBot,
    PlayRandom,
    Exit,
}

#[derive(Debug, Clone)]
struct PlayModeDescription {
    name: &'static str,
    value: SelectPlayModeResult,
}

const AVAILABLE_PLAY_MODES: [PlayModeDescription; 3] = [
    PlayModeDescription {
        name: "Play against bot (offline)",
        value: SelectPlayModeResult::PlayAgainstBot,
    },
    PlayModeDescription {
        name: "Play against random player (online)",
        value: SelectPlayModeResult::PlayRandom,
    },
    PlayModeDescription {
        name: "Exit app",
        value: SelectPlayModeResult::Exit,
    },
];

#[derive(Default)]
struct SelectPlayModeScreen {
    state: ListState,
    curr_selection: usize,
}

pub fn select_play_mode<B: Backend>(
    terminal: &mut Terminal<B>,
) -> io::Result<SelectPlayModeResult> {
    let mut model = SelectPlayModeScreen::default();

    let mut last_tick = Instant::now();
    loop {
        model.state.select(Some(model.curr_selection));
        terminal.draw(|f| ui(f, &mut model))?;

        let timeout = TICK_RATE
            .checked_sub(last_tick.elapsed())
            .unwrap_or_else(|| Duration::from_secs(0));

        if crossterm::event::poll(timeout)? {
            if let Event::Key(key) = event::read()? {
                match key.code {
                    KeyCode::Char('q') => return Ok(SelectPlayModeResult::Exit),
                    KeyCode::Enter => return Ok(AVAILABLE_PLAY_MODES[model.curr_selection].value),
                    KeyCode::Down => model.curr_selection += 1,
                    KeyCode::Up => model.curr_selection += AVAILABLE_PLAY_MODES.len() - 1,
                    _ => {}
                }

                model.curr_selection %= AVAILABLE_PLAY_MODES.len();
            }
        }
        if last_tick.elapsed() >= TICK_RATE {
            last_tick = Instant::now();
        }
    }
}

fn ui<B: Backend>(f: &mut Frame<B>, model: &mut SelectPlayModeScreen) {
    let area = centered_rect_size(50, 5, &f.size());

    // Create a List from all list items and highlight the currently selected one
    let items = AVAILABLE_PLAY_MODES
        .iter()
        .map(|mode| ListItem::new(Text::raw(mode.name)))
        .collect::<Vec<_>>();
    let items = List::new(items)
        .block(
            Block::default()
                .title("Select play mode")
                .borders(Borders::ALL),
        )
        .highlight_style(
            Style::default()
                .fg(HIGHLIGHT_COLOR)
                .add_modifier(Modifier::BOLD),
        )
        .highlight_symbol(">> ");

    f.render_stateful_widget(items, area, &mut model.state);
}
